# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import textwrap
import warnings
from typing import Any, Optional
from unittest.mock import Mock, patch

import pytest
from pydantic.v1 import create_model  # This is temporary for defining properties below

import braket.ir as ir
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.ahs.atom_arrangement import AtomArrangement
from braket.ahs.hamiltonian import Hamiltonian
from braket.annealing import Problem, ProblemType
from braket.circuits import Circuit, FreeParameter, Gate, Noise
from braket.circuits.noise_model import GateCriteria, NoiseModel, NoiseModelInstruction
from braket.circuits.serialization import IRType, SerializableProgram
from braket.device_schema import DeviceActionType, DeviceCapabilities
from braket.device_schema.openqasm_device_action_properties import OpenQASMDeviceActionProperties
from braket.devices import LocalSimulator, local_simulator
from braket.ir.openqasm import Program
from braket.simulator import BraketSimulator
from braket.task_result import AnnealingTaskResult, GateModelTaskResult
from braket.task_result.analog_hamiltonian_simulation_task_result_v1 import (
    AnalogHamiltonianSimulationTaskResult,
)
from braket.tasks import AnnealingQuantumTaskResult, GateModelQuantumTaskResult
from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import (
    AnalogHamiltonianSimulationQuantumTaskResult,
)

GATE_MODEL_RESULT = GateModelTaskResult(**{
    "measurements": [[0, 0], [0, 0], [0, 0], [1, 1]],
    "measuredQubits": [0, 1],
    "taskMetadata": {
        "braketSchemaHeader": {"name": "braket.task_result.task_metadata", "version": "1"},
        "id": "task_arn",
        "shots": 100,
        "deviceId": "default",
    },
    "additionalMetadata": {
        "action": {
            "braketSchemaHeader": {"name": "braket.ir.jaqcd.program", "version": "1"},
            "instructions": [{"control": 0, "target": 1, "type": "cnot"}],
        },
    },
})

ANNEALING_RESULT = AnnealingTaskResult(**{
    "solutions": [[-1, -1, -1, -1], [1, -1, 1, 1], [1, -1, -1, 1]],
    "solutionCounts": [3, 2, 4],
    "values": [0.0, 1.0, 2.0],
    "variableCount": 4,
    "taskMetadata": {
        "id": "task_arn",
        "shots": 100,
        "deviceId": "device_id",
    },
    "additionalMetadata": {
        "action": {
            "type": "ISING",
            "linear": {"0": 0.3333, "1": -0.333, "4": -0.333, "5": 0.333},
            "quadratic": {"0,4": 0.667, "0,5": -1.0, "1,4": 0.667, "1,5": 0.667},
        },
        "dwaveMetadata": {
            "activeVariables": [0],
            "timing": {
                "qpuSamplingTime": 100,
                "qpuAnnealTimePerSample": 20,
                "qpuAccessTime": 10917,
                "qpuAccessOverheadTime": 3382,
                "qpuReadoutTimePerSample": 274,
                "qpuProgrammingTime": 9342,
                "qpuDelayTimePerSample": 21,
                "postProcessingOverheadTime": 117,
                "totalPostProcessingTime": 117,
                "totalRealTime": 10917,
                "runTimeChip": 1575,
                "annealTimePerRun": 20,
                "readoutTimePerRun": 274,
            },
        },
    },
})

AHS_RESULT = AnalogHamiltonianSimulationTaskResult(**{
    "taskMetadata": {
        "id": "rydberg",
        "shots": 100,
        "deviceId": "rydbergLocalSimulator",
    },
})


class DummyCircuitSimulator(BraketSimulator):
    def run(
        self,
        program: ir.jaqcd.Program,
        qubits: int,
        shots: Optional[int],
        inputs: Optional[dict[str, float]],
        *args,
        **kwargs,
    ) -> dict[str, Any]:
        self._shots = shots
        self._qubits = qubits
        return GATE_MODEL_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        return DeviceCapabilities.parse_obj({
            "service": {
                "executionWindows": [
                    {
                        "executionDay": "Everyday",
                        "windowStartHour": "11:00",
                        "windowEndHour": "12:00",
                    }
                ],
                "shotsRange": [1, 10],
            },
            "action": {
                "braket.ir.openqasm.program": {
                    "actionType": "braket.ir.openqasm.program",
                    "version": ["1"],
                },
                "braket.ir.jaqcd.program": {
                    "actionType": "braket.ir.jaqcd.program",
                    "version": ["1"],
                },
            },
            "deviceParameters": {},
        })


class DummyJaqcdSimulator(BraketSimulator):
    def run(
        self,
        program: ir.jaqcd.Program,
        qubits: Optional[int] = None,
        shots: Optional[int] = None,
        *args,
        **kwargs,
    ) -> dict[str, Any]:
        if not isinstance(program, ir.jaqcd.Program):
            raise TypeError("Not a Jaqcd program")
        self._shots = shots
        self._qubits = qubits
        return GATE_MODEL_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        return DeviceCapabilities.parse_obj({
            "service": {
                "executionWindows": [
                    {
                        "executionDay": "Everyday",
                        "windowStartHour": "11:00",
                        "windowEndHour": "12:00",
                    }
                ],
                "shotsRange": [1, 10],
            },
            "action": {
                "braket.ir.jaqcd.program": {
                    "actionType": "braket.ir.jaqcd.program",
                    "version": ["1"],
                },
            },
            "deviceParameters": {},
        })

    def assert_shots(self, shots):
        assert self._shots == shots

    def assert_qubits(self, qubits):
        assert self._qubits == qubits


class DummyProgramSimulator(BraketSimulator):
    def run(
        self,
        openqasm_ir: Program,
        shots: int = 0,
        batch_size: int = 1,
    ) -> GateModelTaskResult:
        return GATE_MODEL_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        device_properties = DeviceCapabilities.parse_obj({
            "service": {
                "executionWindows": [
                    {
                        "executionDay": "Everyday",
                        "windowStartHour": "00:00",
                        "windowEndHour": "23:59:59",
                    }
                ],
                "shotsRange": [1, 10],
            },
            "action": {
                "braket.ir.openqasm.program": {
                    "actionType": "braket.ir.openqasm.program",
                    "version": ["1"],
                }
            },
            "deviceParameters": {},
        })
        oq3_action = OpenQASMDeviceActionProperties.parse_raw(
            json.dumps({
                "actionType": "braket.ir.openqasm.program",
                "version": ["1"],
                "supportedOperations": ["rx", "ry", "h", "cy", "cnot", "unitary"],
                "supportedResultTypes": [
                    {"name": "StateVector", "observables": None, "minShots": 0, "maxShots": 0},
                ],
                "supportedPragmas": [
                    "braket_unitary_matrix",
                    "braket_result_type_sample",
                    "braket_result_type_expectation",
                    "braket_result_type_variance",
                    "braket_result_type_probability",
                    "braket_result_type_state_vector",
                ],
            })
        )
        device_properties.action[DeviceActionType.OPENQASM] = oq3_action
        return device_properties


class DummySerializableProgram(SerializableProgram):
    def __init__(self, source: str):
        self.source = source

    def to_ir(self, ir_type: IRType = IRType.OPENQASM) -> str:
        return self.source


class DummySerializableProgramSimulator(DummyProgramSimulator):
    def run(
        self,
        program: SerializableProgram,
        shots: int = 0,
        batch_size: int = 1,
    ) -> GateModelQuantumTaskResult:
        return GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


class DummyProgramDensityMatrixSimulator(BraketSimulator):
    def run(
        self, program: ir.openqasm.Program, shots: Optional[int], *args, **kwargs
    ) -> dict[str, Any]:
        self._shots = shots
        return GATE_MODEL_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        device_properties = DeviceCapabilities.parse_obj({
            "service": {
                "executionWindows": [
                    {
                        "executionDay": "Everyday",
                        "windowStartHour": "11:00",
                        "windowEndHour": "12:00",
                    }
                ],
                "shotsRange": [1, 10],
            },
            "action": {},
            "deviceParameters": {},
        })
        oq3_action = OpenQASMDeviceActionProperties.parse_raw(
            json.dumps({
                "actionType": "braket.ir.openqasm.program",
                "version": ["1"],
                "supportedOperations": ["rx", "ry", "h", "cy", "cnot", "unitary"],
                "supportedResultTypes": [
                    {"name": "StateVector", "observables": None, "minShots": 0, "maxShots": 0},
                ],
                "supportedPragmas": [
                    "braket_noise_bit_flip",
                    "braket_noise_depolarizing",
                    "braket_noise_kraus",
                    "braket_noise_pauli_channel",
                    "braket_noise_generalized_amplitude_damping",
                    "braket_noise_amplitude_damping",
                    "braket_noise_phase_flip",
                    "braket_noise_phase_damping",
                    "braket_noise_two_qubit_dephasing",
                    "braket_noise_two_qubit_depolarizing",
                    "braket_unitary_matrix",
                    "braket_result_type_sample",
                    "braket_result_type_expectation",
                    "braket_result_type_variance",
                    "braket_result_type_probability",
                    "braket_result_type_density_matrix",
                ],
            })
        )
        device_properties.action[DeviceActionType.OPENQASM] = oq3_action
        return device_properties


class DummyAnnealingSimulator(BraketSimulator):
    def run(self, problem: ir.annealing.Problem, *args, **kwargs) -> AnnealingTaskResult:
        return ANNEALING_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        return DeviceCapabilities.parse_obj({
            "service": {
                "executionWindows": [
                    {
                        "executionDay": "Everyday",
                        "windowStartHour": "11:00",
                        "windowEndHour": "12:00",
                    }
                ],
                "shotsRange": [1, 10],
            },
            "action": {
                "braket.ir.annealing.problem": {
                    "actionType": "braket.ir.annealing.problem",
                    "version": ["1"],
                }
            },
            "deviceParameters": {},
        })


class DummyRydbergSimulator(BraketSimulator):
    def run(
        self, program: AnalogHamiltonianSimulation, *args, **kwargs
    ) -> AnalogHamiltonianSimulationTaskResult:
        return AHS_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        properties = {
            "service": {
                "executionWindows": [
                    {
                        "executionDay": "Everyday",
                        "windowStartHour": "00:00",
                        "windowEndHour": "23:59:59",
                    }
                ],
                "shotsRange": [0, 10],
            },
            "action": {"braket.ir.ahs.program": {}},
        }

        RydbergSimulatorDeviceCapabilities = create_model(
            "RydbergSimulatorDeviceCapabilities", **properties
        )

        return RydbergSimulatorDeviceCapabilities.parse_obj(properties)


mock_circuit_entry = Mock()
mock_program_entry = Mock()
mock_jaqcd_entry = Mock()
mock_circuit_dm_entry = Mock()
mock_circuit_entry.load.return_value = DummyCircuitSimulator
mock_program_entry.load.return_value = DummyProgramSimulator
mock_jaqcd_entry.load.return_value = DummyJaqcdSimulator
mock_circuit_dm_entry.load.return_value = DummyProgramDensityMatrixSimulator
local_simulator._simulator_devices = {
    "dummy": mock_circuit_entry,
    "dummy_oq3": mock_program_entry,
    "dummy_jaqcd": mock_jaqcd_entry,
    "dummy_oq3_dm": mock_circuit_dm_entry,
}

mock_ahs_program = AnalogHamiltonianSimulation(
    register=AtomArrangement(), hamiltonian=Hamiltonian()
)


def test_load_from_entry_point():
    sim = LocalSimulator("dummy_oq3")
    task = sim.run(Circuit().h(0).cnot(0, 1), 10)
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_gate_model():
    dummy = DummyProgramSimulator()
    sim = LocalSimulator(dummy)
    task = sim.run(Circuit().h(0).cnot(0, 1), 10)
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_circuit():
    dummy = DummyProgramSimulator()
    theta = FreeParameter("theta")
    task = Circuit().rx(angle=theta, target=0)
    device = LocalSimulator(dummy)
    num_tasks = 3
    circuits = [task for _ in range(num_tasks)]
    inputs = [{"theta": i} for i in range(num_tasks)]
    batch = device.run_batch(circuits, inputs=inputs, shots=10)
    assert len(batch.results()) == num_tasks
    for x in batch.results():
        assert x == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_with_max_parallel():
    dummy = DummyProgramSimulator()
    task = Circuit().h(0).cnot(0, 1)
    device = LocalSimulator(dummy)
    num_tasks = 3
    circuits = [task for _ in range(num_tasks)]
    batch = device.run_batch(circuits, shots=10, max_parallel=2)
    assert len(batch.results()) == num_tasks
    for x in batch.results():
        assert x == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_with_annealing_problems():
    dummy = DummyAnnealingSimulator()
    problem = Problem(ProblemType.ISING)
    device = LocalSimulator(dummy)
    num_tasks = 3
    problems = [problem for _ in range(num_tasks)]
    batch = device.run_batch(problems, shots=10)
    assert len(batch.results()) == num_tasks
    for x in batch.results():
        assert x == AnnealingQuantumTaskResult.from_object(ANNEALING_RESULT)


def test_batch_circuit_without_inputs():
    dummy = DummyProgramSimulator()
    bell = Circuit().h(0).cnot(0, 1)
    device = LocalSimulator(dummy)
    num_tasks = 3
    circuits = [bell for _ in range(num_tasks)]
    batch = device.run_batch(circuits, shots=10)
    assert len(batch.results()) == num_tasks
    for x in batch.results():
        assert x == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_circuit_with_unbound_parameters():
    dummy = DummyProgramSimulator()
    device = LocalSimulator(dummy)
    theta = FreeParameter("theta")
    task = Circuit().rx(angle=theta, target=0)
    inputs = {"beta": 0.2}
    cannot_execute_with_unbound = "Cannot execute circuit with unbound parameters: {'theta'}"
    with pytest.raises(ValueError, match=cannot_execute_with_unbound):
        device.run_batch(task, inputs=inputs, shots=10)


def test_batch_circuit_with_single_task():
    dummy = DummyProgramSimulator()
    bell = Circuit().h(0).cnot(0, 1)
    device = LocalSimulator(dummy)
    batch = device.run_batch(bell, shots=10)
    assert len(batch.results()) == 1
    assert batch.results()[0] == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_circuit_with_task_and_input_mismatch():
    dummy = DummyProgramSimulator()
    bell = Circuit().h(0).cnot(0, 1)
    device = LocalSimulator(dummy)
    num_tasks = 3
    circuits = [bell for _ in range(num_tasks)]
    inputs = [{} for _ in range(num_tasks - 1)]
    with pytest.raises(ValueError):
        device.run_batch(circuits, inputs=inputs, shots=10)


def test_run_gate_model_inputs():
    dummy = DummyProgramSimulator()
    dummy.run = Mock(return_value=GATE_MODEL_RESULT)
    sim = LocalSimulator(dummy)
    circuit = Circuit().rx(0, FreeParameter("theta"))
    task = sim.run(circuit, inputs={"theta": 2}, shots=10)
    dummy.run.assert_called_with(
        Program(
            source="\n".join((
                "OPENQASM 3.0;",
                "input float theta;",
                "bit[1] b;",
                "qubit[1] q;",
                "rx(theta) q[0];",
                "b[0] = measure q[0];",
            )),
            inputs={"theta": 2},
        ),
        shots=10,
    )
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_program_model_inputs():
    dummy = DummyProgramSimulator()
    dummy.run = Mock(return_value=GATE_MODEL_RESULT)
    sim = LocalSimulator(dummy)
    inputs = {"theta": 2}
    source_string = (
        "OPENQASM 3.0;",
        "input float theta;",
        "bit[1] b;",
        "qubit[1] q;",
        "rx(theta) q[0];",
        "b[0] = measure q[0];",
    )
    program = Program.construct(source="\n".join(source_string), inputs=inputs)
    update_inputs = {"beta": 3}
    task = sim.run(program, inputs=update_inputs, shots=10)
    assert program.inputs == inputs
    program.inputs.update(update_inputs)
    dummy.run.assert_called_with(program, shots=10)
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_jaqcd_only():
    dummy = DummyJaqcdSimulator()
    sim = LocalSimulator(dummy)
    task = sim.run(Circuit().h(0).cnot(0, 1), 10)
    dummy.assert_shots(10)
    dummy.assert_qubits(None)
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_program_model():
    dummy = DummyProgramSimulator()
    sim = LocalSimulator(dummy)
    task = sim.run(
        Program(
            source="""
qubit[2] q;
bit[2] c;

h q[0];
cnot q[0], q[1];

c = measure q;
"""
        )
    )
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_serializable_program_model():
    dummy = DummySerializableProgramSimulator()
    sim = LocalSimulator(dummy)
    task = sim.run(
        DummySerializableProgram(
            source="""
qubit[2] q;
bit[2] c;
h q[0];
cnot q[0], q[1];
c = measure q;
"""
        )
    )
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_serializable_program_model_with_inputs():
    dummy = DummySerializableProgramSimulator()
    sim = LocalSimulator(dummy)
    task = sim.run(
        DummySerializableProgram(
            source="""
input float a;
qubit[2] q;
bit[2] c;
h q[0];
cnot q[0], q[1];
c = measure q;
"""
        ),
        inputs={"a": 0.1},
    )
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


@pytest.mark.xfail(raises=ValueError)
def test_run_gate_model_value_error():
    dummy = DummyCircuitSimulator()
    sim = LocalSimulator(dummy)
    sim.run(Circuit().h(0).cnot(0, 1))


def test_run_annealing():
    sim = LocalSimulator(DummyAnnealingSimulator())
    task = sim.run(Problem(ProblemType.ISING))
    assert task.result() == AnnealingQuantumTaskResult.from_object(ANNEALING_RESULT)


def test_run_ahs():
    sim = LocalSimulator(DummyRydbergSimulator())
    task = sim.run(mock_ahs_program)
    assert task.result() == AnalogHamiltonianSimulationQuantumTaskResult.from_object(AHS_RESULT)

    task = sim.run(mock_ahs_program.to_ir())
    assert task.result() == AnalogHamiltonianSimulationQuantumTaskResult.from_object(AHS_RESULT)


def test_registered_backends():
    assert LocalSimulator.registered_backends() == {
        "dummy",
        "dummy_oq3",
        "dummy_jaqcd",
        "dummy_oq3_dm",
    }


@pytest.mark.xfail(raises=TypeError)
def test_init_invalid_backend_type():
    LocalSimulator(1234)


@pytest.mark.xfail(raises=ValueError)
def test_init_unregistered_backend():
    LocalSimulator("foo")


@pytest.mark.xfail(raises=NotImplementedError)
def test_run_unsupported_type():
    sim = LocalSimulator(DummyCircuitSimulator())
    sim.run("I'm unsupported")


@pytest.mark.xfail(raises=NotImplementedError)
def test_run_annealing_unsupported():
    sim = LocalSimulator(DummyCircuitSimulator())
    sim.run(Problem(ProblemType.ISING))


@pytest.mark.xfail(raises=NotImplementedError)
def test_run_qubit_gate_unsupported():
    sim = LocalSimulator(DummyAnnealingSimulator())
    sim.run(Circuit().h(0).cnot(0, 1), 1000)


def test_properties():
    dummy = DummyCircuitSimulator()
    sim = LocalSimulator(dummy)
    expected_properties = dummy.properties
    assert sim.properties == expected_properties


@pytest.fixture
def noise_model():
    return (
        NoiseModel()
        .add_noise(Noise.BitFlip(0.05), GateCriteria(Gate.H))
        .add_noise(Noise.TwoQubitDepolarizing(0.10), GateCriteria(Gate.CNot))
    )


@pytest.mark.parametrize("backend", ["dummy_oq3_dm"])
def test_valid_local_device_for_noise_model(backend, noise_model):
    device = LocalSimulator(backend, noise_model=noise_model)
    assert device._noise_model.instructions == [
        NoiseModelInstruction(Noise.BitFlip(0.05), GateCriteria(Gate.H)),
        NoiseModelInstruction(Noise.TwoQubitDepolarizing(0.10), GateCriteria(Gate.CNot)),
    ]


@pytest.mark.parametrize("backend", ["dummy_oq3"])
def test_invalid_local_device_for_noise_model(backend, noise_model):
    with pytest.raises(ValueError):
        _ = LocalSimulator(backend, noise_model=noise_model)


@pytest.mark.parametrize("backend", ["dummy_oq3_dm"])
def test_local_device_with_invalid_noise_model(backend, noise_model):
    with pytest.raises(TypeError):
        _ = LocalSimulator(backend, noise_model=Mock())


@patch.object(DummyProgramDensityMatrixSimulator, "run")
def test_run_with_noise_model(mock_run, noise_model):
    mock_run.return_value = GATE_MODEL_RESULT
    device = LocalSimulator("dummy_oq3_dm", noise_model=noise_model)
    circuit = Circuit().h(0).cnot(0, 1)
    _ = device.run(circuit, shots=4)

    expected_circuit = textwrap.dedent(
        """
        OPENQASM 3.0;
        bit[2] b;
        qubit[2] q;
        h q[0];
        #pragma braket noise bit_flip(0.05) q[0]
        cnot q[0], q[1];
        #pragma braket noise two_qubit_depolarizing(0.1) q[0], q[1]
        b[0] = measure q[0];
        b[1] = measure q[1];
        """
    ).strip()

    mock_run.assert_called_with(
        Program(source=expected_circuit, inputs={}),
        shots=4,
    )


@patch.object(LocalSimulator, "_apply_noise_model_to_circuit")
def test_run_batch_with_noise_model(mock_apply, noise_model):
    device = LocalSimulator("dummy_oq3_dm", noise_model=noise_model)
    circuit = Circuit().h(0).cnot(0, 1)

    mock_apply.return_value = noise_model.apply(circuit)
    _ = device.run_batch([circuit] * 2, shots=4).results()
    assert mock_apply.call_count == 2


@patch.object(DummyProgramDensityMatrixSimulator, "run")
def test_run_noisy_circuit_with_noise_model(mock_run, noise_model):
    mock_run.return_value = GATE_MODEL_RESULT
    device = LocalSimulator("dummy_oq3_dm", noise_model=noise_model)
    circuit = Circuit().h(0).depolarizing(0, 0.1)
    with warnings.catch_warnings(record=True) as w:
        _ = device.run(circuit, shots=4)

    expected_warning = (
        "The noise model of the device is applied to a circuit that already has noise instructions."
    )
    expected_circuit = textwrap.dedent(
        """
        OPENQASM 3.0;
        bit[1] b;
        qubit[1] q;
        h q[0];
        #pragma braket noise bit_flip(0.05) q[0]
        #pragma braket noise depolarizing(0.1) q[0]
        b[0] = measure q[0];
        """
    ).strip()

    mock_run.assert_called_with(
        Program(source=expected_circuit, inputs={}),
        shots=4,
    )
    assert w[-1].message.__str__() == expected_warning


@patch.object(DummyProgramDensityMatrixSimulator, "run")
def test_run_openqasm_with_noise_model(mock_run, noise_model):
    mock_run.return_value = GATE_MODEL_RESULT
    device = LocalSimulator("dummy_oq3_dm", noise_model=noise_model)
    expected_circuit = textwrap.dedent(
        """
        OPENQASM 3.0;
        bit[1] b;
        qubit[1] q;
        h q[0];
        b[0] = measure q[0];
        """
    ).strip()
    expected_warning = (
        "Noise model is only applicable to circuits. The type of the task specification "
        "is Program. The noise model of the device does not apply."
    )
    circuit = Program(source=expected_circuit)
    with warnings.catch_warnings(record=True) as w:
        _ = device.run(circuit, shots=4)

    mock_run.assert_called_with(
        Program(source=expected_circuit, inputs=None),
        shots=4,
    )
    assert w[-1].message.__str__() == expected_warning



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import random
import string
from abc import ABC, abstractmethod

import numpy as np
import scipy as sp
from oqpy import WaveformVar, bool_, complex128, declare_waveform_generator, duration, float64
from oqpy.base import OQPyExpression

from braket.parametric.free_parameter import FreeParameter
from braket.parametric.free_parameter_expression import (
    FreeParameterExpression,
    subs_if_free_parameter,
)
from braket.parametric.parameterizable import Parameterizable


class Waveform(ABC):
    """A waveform is a time-dependent envelope that can be used to emit signals on an output port
    or receive signals from an input port. As such, when transmitting signals to the qubit, a
    frame determines time at which the waveform envelope is emitted, its carrier frequency, and
    it's phase offset. When capturing signals from a qubit, at minimum a frame determines the
    time at which the signal is captured. See https://openqasm.com/language/openpulse.html#waveforms
    for more details.
    """

    @abstractmethod
    def _to_oqpy_expression(self) -> OQPyExpression:
        """Returns an OQPyExpression defining this waveform."""

    @abstractmethod
    def sample(self, dt: float) -> np.ndarray:
        """Generates a sample of amplitudes for this Waveform based on the given time resolution.

        Args:
            dt (float): The time resolution.

        Returns:
            np.ndarray: The sample amplitudes for this waveform.
        """

    @staticmethod
    @abstractmethod
    def _from_calibration_schema(waveform_json: dict) -> Waveform:
        """Parses a JSON input and returns the BDK waveform. See https://github.com/aws/amazon-braket-schemas-python/blob/main/src/braket/device_schema/pulse/native_gate_calibrations_v1.py#L104

        Args:
            waveform_json (dict): A JSON object with the needed parameters for making the Waveform.

        Returns:
            Waveform: A Waveform object parsed from the supplied JSON.
        """


class ArbitraryWaveform(Waveform):
    """An arbitrary waveform with amplitudes at each timestep explicitly specified using
    an array.
    """

    def __init__(self, amplitudes: list[complex], id: str | None = None):
        """Initializes an `ArbitraryWaveform`.

        Args:
            amplitudes (list[complex]): Array of complex values specifying the
                waveform amplitude at each timestep. The timestep is determined by the sampling rate
                of the frame to which waveform is applied to.
            id (Optional[str]): The identifier used for declaring this waveform. A random string of
                ascii characters is assigned by default.
        """
        self.amplitudes = list(amplitudes)
        self.id = id or _make_identifier_name()

    def __repr__(self) -> str:
        return f"ArbitraryWaveform('id': {self.id}, 'amplitudes': {self.amplitudes})"

    def __eq__(self, other: ArbitraryWaveform):
        return isinstance(other, ArbitraryWaveform) and (self.amplitudes, self.id) == (
            other.amplitudes,
            other.id,
        )

    def _to_oqpy_expression(self) -> OQPyExpression:
        """Returns an OQPyExpression defining this waveform.

        Returns:
            OQPyExpression: The OQPyExpression.
        """
        return WaveformVar(init_expression=self.amplitudes, name=self.id)

    def sample(self, dt: float) -> np.ndarray:
        """Generates a sample of amplitudes for this Waveform based on the given time resolution.

        Args:
            dt (float): The time resolution.

        Raises:
            NotImplementedError: This class does not implement sample.

        Returns:
            np.ndarray: The sample amplitudes for this waveform.
        """
        raise NotImplementedError

    @staticmethod
    def _from_calibration_schema(waveform_json: dict) -> ArbitraryWaveform:
        wave_id = waveform_json["waveformId"]
        complex_amplitudes = [complex(i[0], i[1]) for i in waveform_json["amplitudes"]]
        return ArbitraryWaveform(complex_amplitudes, wave_id)


class ConstantWaveform(Waveform, Parameterizable):
    """A constant waveform which holds the supplied `iq` value as its amplitude for the
    specified length.
    """

    def __init__(self, length: float | FreeParameterExpression, iq: complex, id: str | None = None):
        """Initializes a `ConstantWaveform`.

        Args:
            length (Union[float, FreeParameterExpression]): Value (in seconds)
                specifying the duration of the waveform.
            iq (complex): complex value specifying the amplitude of the waveform.
            id (Optional[str]): The identifier used for declaring this waveform. A random string of
                ascii characters is assigned by default.
        """
        self.length = length
        self.iq = iq
        self.id = id or _make_identifier_name()

    def __repr__(self) -> str:
        return f"ConstantWaveform('id': {self.id}, 'length': {self.length}, 'iq': {self.iq})"

    @property
    def parameters(self) -> list[FreeParameterExpression | FreeParameter | float]:
        """Returns the parameters associated with the object, either unbound free parameter
        expressions or bound values.

        Returns:
            list[Union[FreeParameterExpression, FreeParameter, float]]: a list of parameters.
        """
        return [self.length]

    def bind_values(self, **kwargs: FreeParameter | str) -> ConstantWaveform:
        """Takes in parameters and returns an object with specified parameters
        replaced with their values.

        Args:
            **kwargs (Union[FreeParameter, str]): Arbitrary keyword arguments.

        Returns:
            ConstantWaveform: A copy of this waveform with the requested parameters bound.
        """
        constructor_kwargs = {
            "length": subs_if_free_parameter(self.length, **kwargs),
            "iq": self.iq,
            "id": self.id,
        }
        return ConstantWaveform(**constructor_kwargs)

    def __eq__(self, other: ConstantWaveform):
        return isinstance(other, ConstantWaveform) and (self.length, self.iq, self.id) == (
            other.length,
            other.iq,
            other.id,
        )

    def _to_oqpy_expression(self) -> OQPyExpression:
        """Returns an OQPyExpression defining this waveform.

        Returns:
            OQPyExpression: The OQPyExpression.
        """
        constant_generator = declare_waveform_generator(
            "constant", [("length", duration), ("iq", complex128)]
        )
        return WaveformVar(
            init_expression=constant_generator(self.length, self.iq),
            name=self.id,
        )

    def sample(self, dt: float) -> np.ndarray:
        """Generates a sample of amplitudes for this Waveform based on the given time resolution.

        Args:
            dt (float): The time resolution.

        Returns:
            np.ndarray: The sample amplitudes for this waveform.
        """
        # Amplitudes should be gated by [0:self.length]
        sample_range = np.arange(0, self.length, dt)
        return self.iq * np.ones_like(sample_range)

    @staticmethod
    def _from_calibration_schema(waveform_json: dict) -> ConstantWaveform:
        wave_id = waveform_json["waveformId"]
        length = iq = None
        for val in waveform_json["arguments"]:
            if val["name"] == "length":
                length = (
                    float(val["value"])
                    if val["type"] == "float"
                    else FreeParameterExpression(val["value"])
                )
            if val["name"] == "iq":
                iq = (
                    complex(val["value"])
                    if val["type"] == "complex"
                    else FreeParameterExpression(val["value"])
                )
        return ConstantWaveform(length=length, iq=iq, id=wave_id)


class DragGaussianWaveform(Waveform, Parameterizable):
    """A gaussian waveform with an additional gaussian derivative component and lifting applied."""

    def __init__(
        self,
        length: float | FreeParameterExpression,
        sigma: float | FreeParameterExpression,
        beta: float | FreeParameterExpression,
        amplitude: float | FreeParameterExpression = 1,
        zero_at_edges: bool = False,
        id: str | None = None,
    ):
        """Initializes a `DragGaussianWaveform`.

        Args:
            length (Union[float, FreeParameterExpression]): Value (in seconds)
                specifying the duration of the waveform.
            sigma (Union[float, FreeParameterExpression]): A measure (in seconds) of
                how wide or narrow the Gaussian peak is.
            beta (Union[float, FreeParameterExpression]): The correction amplitude.
            amplitude (Union[float, FreeParameterExpression]): The amplitude of the
                waveform envelope. Defaults to 1.
            zero_at_edges (bool): bool specifying whether the waveform amplitude is clipped to
                zero at the edges. Defaults to False.
            id (Optional[str]): The identifier used for declaring this waveform. A random string of
                ascii characters is assigned by default.
        """
        self.length = length
        self.sigma = sigma
        self.beta = beta
        self.amplitude = amplitude
        self.zero_at_edges = zero_at_edges
        self.id = id or _make_identifier_name()

    def __repr__(self) -> str:
        return (
            f"DragGaussianWaveform('id': {self.id}, 'length': {self.length}, "
            f"'sigma': {self.sigma}, 'beta': {self.beta}, 'amplitude': {self.amplitude}, "
            f"'zero_at_edges': {self.zero_at_edges})"
        )

    @property
    def parameters(self) -> list[FreeParameterExpression | FreeParameter | float]:
        """Returns the parameters associated with the object, either unbound free parameter
        expressions or bound values.
        """
        return [self.length, self.sigma, self.beta, self.amplitude]

    def bind_values(self, **kwargs: FreeParameter | str) -> DragGaussianWaveform:
        """Takes in parameters and returns an object with specified parameters
        replaced with their values.

        Args:
            **kwargs (Union[FreeParameter, str]): Arbitrary keyword arguments.

        Returns:
            DragGaussianWaveform: A copy of this waveform with the requested parameters bound.
        """
        constructor_kwargs = {
            "length": subs_if_free_parameter(self.length, **kwargs),
            "sigma": subs_if_free_parameter(self.sigma, **kwargs),
            "beta": subs_if_free_parameter(self.beta, **kwargs),
            "amplitude": subs_if_free_parameter(self.amplitude, **kwargs),
            "zero_at_edges": self.zero_at_edges,
            "id": self.id,
        }
        return DragGaussianWaveform(**constructor_kwargs)

    def __eq__(self, other: DragGaussianWaveform):
        return isinstance(other, DragGaussianWaveform) and (
            self.length,
            self.sigma,
            self.beta,
            self.amplitude,
            self.zero_at_edges,
            self.id,
        ) == (other.length, other.sigma, other.beta, other.amplitude, other.zero_at_edges, other.id)

    def _to_oqpy_expression(self) -> OQPyExpression:
        """Returns an OQPyExpression defining this waveform.

        Returns:
            OQPyExpression: The OQPyExpression.
        """
        drag_gaussian_generator = declare_waveform_generator(
            "drag_gaussian",
            [
                ("length", duration),
                ("sigma", duration),
                ("beta", float64),
                ("amplitude", float64),
                ("zero_at_edges", bool_),
            ],
        )
        return WaveformVar(
            init_expression=drag_gaussian_generator(
                self.length,
                self.sigma,
                self.beta,
                self.amplitude,
                self.zero_at_edges,
            ),
            name=self.id,
        )

    def sample(self, dt: float) -> np.ndarray:
        """Generates a sample of amplitudes for this Waveform based on the given time resolution.

        Args:
            dt (float): The time resolution.

        Returns:
            np.ndarray: The sample amplitudes for this waveform.
        """
        sample_range = np.arange(0, self.length, dt)
        t0 = self.length / 2
        zero_at_edges_int = int(self.zero_at_edges)
        return (
            (1 - (1.0j * self.beta * ((sample_range - t0) / self.sigma**2)))
            * (
                self.amplitude
                / (1 - zero_at_edges_int * np.exp(-0.5 * ((self.length / (2 * self.sigma)) ** 2)))
            )
            * (
                np.exp(-0.5 * (((sample_range - t0) / self.sigma) ** 2))
                - zero_at_edges_int * np.exp(-0.5 * ((self.length / (2 * self.sigma)) ** 2))
            )
        )

    @staticmethod
    def _from_calibration_schema(waveform_json: dict) -> DragGaussianWaveform:
        waveform_parameters = {"id": waveform_json["waveformId"]}
        for val in waveform_json["arguments"]:
            waveform_parameters[val["name"]] = (
                float(val["value"])
                if val["type"] == "float"
                else FreeParameterExpression(val["value"])
            )
        return DragGaussianWaveform(**waveform_parameters)


class GaussianWaveform(Waveform, Parameterizable):
    """A waveform with amplitudes following a gaussian distribution for the specified parameters."""

    def __init__(
        self,
        length: float | FreeParameterExpression,
        sigma: float | FreeParameterExpression,
        amplitude: float | FreeParameterExpression = 1,
        zero_at_edges: bool = False,
        id: str | None = None,
    ):
        """Initializes a `GaussianWaveform`.

        Args:
            length (Union[float, FreeParameterExpression]): Value (in seconds) specifying the
                duration of the waveform.
            sigma (Union[float, FreeParameterExpression]): A measure (in seconds) of how wide
                or narrow the Gaussian peak is.
            amplitude (Union[float, FreeParameterExpression]): The amplitude of the waveform
                envelope. Defaults to 1.
            zero_at_edges (bool): bool specifying whether the waveform amplitude is clipped to
                zero at the edges. Defaults to False.
            id (Optional[str]): The identifier used for declaring this waveform. A random string of
                ascii characters is assigned by default.
        """
        self.length = length
        self.sigma = sigma
        self.amplitude = amplitude
        self.zero_at_edges = zero_at_edges
        self.id = id or _make_identifier_name()

    def __repr__(self) -> str:
        return (
            f"GaussianWaveform('id': {self.id}, 'length': {self.length}, 'sigma': {self.sigma}, "
            f"'amplitude': {self.amplitude}, 'zero_at_edges': {self.zero_at_edges})"
        )

    @property
    def parameters(self) -> list[FreeParameterExpression | FreeParameter | float]:
        """Returns the parameters associated with the object, either unbound free parameter
        expressions or bound values.
        """
        return [self.length, self.sigma, self.amplitude]

    def bind_values(self, **kwargs: FreeParameter | str) -> GaussianWaveform:
        """Takes in parameters and returns an object with specified parameters
        replaced with their values.

        Args:
            **kwargs (Union[FreeParameter, str]): Arbitrary keyword arguments.

        Returns:
            GaussianWaveform: A copy of this waveform with the requested parameters bound.
        """
        constructor_kwargs = {
            "length": subs_if_free_parameter(self.length, **kwargs),
            "sigma": subs_if_free_parameter(self.sigma, **kwargs),
            "amplitude": subs_if_free_parameter(self.amplitude, **kwargs),
            "zero_at_edges": self.zero_at_edges,
            "id": self.id,
        }
        return GaussianWaveform(**constructor_kwargs)

    def __eq__(self, other: GaussianWaveform):
        return isinstance(other, GaussianWaveform) and (
            self.length,
            self.sigma,
            self.amplitude,
            self.zero_at_edges,
            self.id,
        ) == (other.length, other.sigma, other.amplitude, other.zero_at_edges, other.id)

    def _to_oqpy_expression(self) -> OQPyExpression:
        """Returns an OQPyExpression defining this waveform.

        Returns:
            OQPyExpression: The OQPyExpression.
        """
        gaussian_generator = declare_waveform_generator(
            "gaussian",
            [
                ("length", duration),
                ("sigma", duration),
                ("amplitude", float64),
                ("zero_at_edges", bool_),
            ],
        )
        return WaveformVar(
            init_expression=gaussian_generator(
                self.length,
                self.sigma,
                self.amplitude,
                self.zero_at_edges,
            ),
            name=self.id,
        )

    def sample(self, dt: float) -> np.ndarray:
        """Generates a sample of amplitudes for this Waveform based on the given time resolution.

        Args:
            dt (float): The time resolution.

        Returns:
            np.ndarray: The sample amplitudes for this waveform.
        """
        sample_range = np.arange(0, self.length, dt)
        t0 = self.length / 2
        zero_at_edges_int = int(self.zero_at_edges)
        return (
            self.amplitude
            / (1 - zero_at_edges_int * np.exp(-0.5 * ((self.length / (2 * self.sigma)) ** 2)))
        ) * (
            np.exp(-0.5 * (((sample_range - t0) / self.sigma) ** 2))
            - zero_at_edges_int * np.exp(-0.5 * ((self.length / (2 * self.sigma)) ** 2))
        )

    @staticmethod
    def _from_calibration_schema(waveform_json: dict) -> GaussianWaveform:
        waveform_parameters = {"id": waveform_json["waveformId"]}
        for val in waveform_json["arguments"]:
            waveform_parameters[val["name"]] = (
                float(val["value"])
                if val["type"] == "float"
                else FreeParameterExpression(val["value"])
            )
        return GaussianWaveform(**waveform_parameters)


class ErfSquareWaveform(Waveform, Parameterizable):
    """A square waveform with smoothed edges."""

    def __init__(
        self,
        length: float | FreeParameterExpression,
        width: float | FreeParameterExpression,
        sigma: float | FreeParameterExpression,
        off_center: float | FreeParameterExpression = 0,
        amplitude: float | FreeParameterExpression = 1,
        zero_at_edges: bool = False,
        id: str | None = None,
    ):
        r"""Initializes a `ErfSquareWaveform`.

        .. math:: (\text{step}((t-t_1)/sigma) + \text{step}(-(t-t_2)/sigma) - 1)

        where :math:`\text{step}(t)` is the rounded step function defined as
        :math:`(erf(t)+1)/2` and :math:`t_1` and :math:`t_2` are the timestamps at the half
        height. The waveform is scaled such that its maximum is equal to `amplitude`.

        Args:
            length (float | FreeParameterExpression): Duration (in seconds) from the start
                to the end of the waveform.
            width (float | FreeParameterExpression): Duration (in seconds) between the
                half height of the two edges.
            sigma (float | FreeParameterExpression): A characteristic time of how quickly
                the edges rise and fall.
            off_center (float | FreeParameterExpression): Shift the smoothed square waveform
                earlier or later in time. When positive, the smoothed square is shifted later
                (to the right), otherwise earlier (to the left). Defaults to 0.
            amplitude (float | FreeParameterExpression): The amplitude of the waveform
                envelope. Defaults to 1.
            zero_at_edges (bool): Whether the waveform is scaled such that it has zero value at the
                edges. Defaults to False.
            id (str | None): The identifier used for declaring this waveform. A random string of
                ascii characters is assigned by default.
        """
        self.length = length
        self.width = width
        self.sigma = sigma
        self.off_center = off_center
        self.amplitude = amplitude
        self.zero_at_edges = zero_at_edges
        self.id = id or _make_identifier_name()

    def __repr__(self) -> str:
        return (
            f"ErfSquareWaveform('id': {self.id}, 'length': {self.length}, "
            f"'width': {self.width}, 'sigma': {self.sigma}, 'off_center': {self.off_center}, "
            f"'amplitude': {self.amplitude}, 'zero_at_edges': {self.zero_at_edges})"
        )

    @property
    def parameters(self) -> list[FreeParameterExpression | FreeParameter | float]:
        """Returns the parameters associated with the object, either unbound free parameter
        expressions or bound values.
        """
        return [self.length, self.width, self.sigma, self.off_center, self.amplitude]

    def bind_values(self, **kwargs: FreeParameter | str) -> ErfSquareWaveform:
        """Takes in parameters and returns an object with specified parameters
        replaced with their values.

        Args:
            **kwargs (FreeParameter | str): Arbitrary keyword arguments.

        Returns:
            ErfSquareWaveform: A copy of this waveform with the requested parameters bound.
        """
        constructor_kwargs = {
            "length": subs_if_free_parameter(self.length, **kwargs),
            "width": subs_if_free_parameter(self.width, **kwargs),
            "sigma": subs_if_free_parameter(self.sigma, **kwargs),
            "off_center": subs_if_free_parameter(self.off_center, **kwargs),
            "amplitude": subs_if_free_parameter(self.amplitude, **kwargs),
            "zero_at_edges": self.zero_at_edges,
            "id": self.id,
        }
        return ErfSquareWaveform(**constructor_kwargs)

    def __eq__(self, other: ErfSquareWaveform):
        return isinstance(other, ErfSquareWaveform) and (
            self.length,
            self.width,
            self.sigma,
            self.off_center,
            self.amplitude,
            self.zero_at_edges,
            self.id,
        ) == (
            other.length,
            other.width,
            other.sigma,
            other.off_center,
            other.amplitude,
            other.zero_at_edges,
            other.id,
        )

    def _to_oqpy_expression(self) -> OQPyExpression:
        """Returns an OQPyExpression defining this waveform.

        Returns:
            OQPyExpression: The OQPyExpression.
        """
        erf_square_generator = declare_waveform_generator(
            "erf_square",
            [
                ("length", duration),
                ("width", duration),
                ("sigma", duration),
                ("off_center", duration),
                ("amplitude", float64),
                ("zero_at_edges", bool_),
            ],
        )
        return WaveformVar(
            init_expression=erf_square_generator(
                self.length,
                self.width,
                self.sigma,
                self.off_center,
                self.amplitude,
                self.zero_at_edges,
            ),
            name=self.id,
        )

    def sample(self, dt: float) -> np.ndarray:
        """Generates a sample of amplitudes for this Waveform based on the given time resolution.

        Args:
            dt (float): The time resolution.

        Returns:
            np.ndarray: The sample amplitudes for this waveform.
        """
        sample_range = np.arange(0, self.length, dt)
        t1 = (self.length - self.width) / 2 + self.off_center
        t2 = (self.length + self.width) / 2 + self.off_center
        samples = (
            sp.special.erf((sample_range - t1) / self.sigma)
            + sp.special.erf(-(sample_range - t2) / self.sigma)
        ) / 2

        mid_waveform_height = sp.special.erf((self.width / 2) / self.sigma)
        waveform_bottom = (sp.special.erf(-t1 / self.sigma) + sp.special.erf(t2 / self.sigma)) / 2

        if self.zero_at_edges:
            return (
                (samples - waveform_bottom)
                / (mid_waveform_height - waveform_bottom)
                * self.amplitude
            )
        return samples * self.amplitude / mid_waveform_height

    @staticmethod
    def _from_calibration_schema(waveform_json: dict) -> ErfSquareWaveform:
        waveform_parameters = {"id": waveform_json["waveformId"]}
        for val in waveform_json["arguments"]:
            waveform_parameters[val["name"]] = (
                float(val["value"])
                if val["type"] == "float"
                else FreeParameterExpression(val["value"])
            )
        return ErfSquareWaveform(**waveform_parameters)


def _make_identifier_name() -> str:
    return "".join([random.choice(string.ascii_letters) for _ in range(10)])  # noqa: S311


def _parse_waveform_from_calibration_schema(waveform: dict) -> Waveform:
    waveform_names = {
        "arbitrary": ArbitraryWaveform._from_calibration_schema,
        "drag_gaussian": DragGaussianWaveform._from_calibration_schema,
        "gaussian": GaussianWaveform._from_calibration_schema,
        "constant": ConstantWaveform._from_calibration_schema,
        "erf_square": ErfSquareWaveform._from_calibration_schema,
    }
    if "amplitudes" in waveform:
        waveform["name"] = "arbitrary"
    if waveform["name"] in waveform_names:
        return waveform_names[waveform["name"]](waveform)
    waveform_id = waveform["waveformId"]
    raise ValueError(f"The waveform {waveform_id} of cannot be constructed")



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import functools

import numpy as np
import pytest

import braket.ir.jaqcd as ir
from braket.circuits import Circuit, FreeParameter, Gate, Instruction, QubitSet
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    QubitReferenceType,
)
from braket.ir.jaqcd.shared_models import (
    Angle,
    DoubleControl,
    DoubleTarget,
    MultiTarget,
    SingleControl,
    SingleTarget,
    TwoDimensionalMatrix,
)
from braket.pulse import ArbitraryWaveform, Frame, Port, PulseSequence


class NoTarget:
    pass


class DoubleAngle:
    pass


class TripleAngle:
    pass


class SingleNegControlModifier:
    pass


testdata = [
    (Gate.H, "h", ir.H, [SingleTarget], {}),
    (Gate.GPhase, "gphase", None, [NoTarget, Angle], {}),
    (Gate.I, "i", ir.I, [SingleTarget], {}),
    (Gate.X, "x", ir.X, [SingleTarget], {}),
    (Gate.Y, "y", ir.Y, [SingleTarget], {}),
    (Gate.Z, "z", ir.Z, [SingleTarget], {}),
    (Gate.S, "s", ir.S, [SingleTarget], {}),
    (Gate.Si, "si", ir.Si, [SingleTarget], {}),
    (Gate.T, "t", ir.T, [SingleTarget], {}),
    (Gate.Ti, "ti", ir.Ti, [SingleTarget], {}),
    (Gate.V, "v", ir.V, [SingleTarget], {}),
    (Gate.Vi, "vi", ir.Vi, [SingleTarget], {}),
    (Gate.Rx, "rx", ir.Rx, [SingleTarget, Angle], {}),
    (Gate.Ry, "ry", ir.Ry, [SingleTarget, Angle], {}),
    (Gate.Rz, "rz", ir.Rz, [SingleTarget, Angle], {}),
    (Gate.U, "u", None, [SingleTarget, TripleAngle], {}),
    (Gate.CNot, "cnot", ir.CNot, [SingleTarget, SingleControl], {}),
    (Gate.CV, "cv", ir.CV, [SingleTarget, SingleControl], {}),
    (Gate.CCNot, "ccnot", ir.CCNot, [SingleTarget, DoubleControl], {}),
    (Gate.Swap, "swap", ir.Swap, [DoubleTarget], {}),
    (Gate.CSwap, "cswap", ir.CSwap, [SingleControl, DoubleTarget], {}),
    (Gate.ISwap, "iswap", ir.ISwap, [DoubleTarget], {}),
    (Gate.PSwap, "pswap", ir.PSwap, [DoubleTarget, Angle], {}),
    (Gate.XY, "xy", ir.XY, [DoubleTarget, Angle], {}),
    (Gate.PhaseShift, "phaseshift", ir.PhaseShift, [SingleTarget, Angle], {}),
    (Gate.CPhaseShift, "cphaseshift", ir.CPhaseShift, [SingleControl, SingleTarget, Angle], {}),
    (
        Gate.CPhaseShift00,
        "cphaseshift00",
        ir.CPhaseShift00,
        [SingleControl, SingleTarget, Angle],
        {},
    ),
    (
        Gate.CPhaseShift01,
        "cphaseshift01",
        ir.CPhaseShift01,
        [SingleControl, SingleTarget, Angle],
        {},
    ),
    (
        Gate.CPhaseShift10,
        "cphaseshift10",
        ir.CPhaseShift10,
        [SingleControl, SingleTarget, Angle],
        {},
    ),
    (Gate.CY, "cy", ir.CY, [SingleTarget, SingleControl], {}),
    (Gate.CZ, "cz", ir.CZ, [SingleTarget, SingleControl], {}),
    (Gate.ECR, "ecr", ir.ECR, [DoubleTarget], {}),
    (Gate.XX, "xx", ir.XX, [DoubleTarget, Angle], {}),
    (Gate.YY, "yy", ir.YY, [DoubleTarget, Angle], {}),
    (Gate.ZZ, "zz", ir.ZZ, [DoubleTarget, Angle], {}),
    (Gate.GPi, "gpi", None, [SingleTarget, Angle], {}),
    (Gate.GPi2, "gpi2", None, [SingleTarget, Angle], {}),
    (Gate.PRx, "prx", None, [SingleTarget, DoubleAngle], {}),
    (Gate.MS, "ms", None, [DoubleTarget, TripleAngle], {}),
    (
        Gate.Unitary,
        "unitary",
        ir.Unitary,
        [TwoDimensionalMatrix, MultiTarget],
        {"input_type": complex},
    ),
    (
        Gate.Unitary,
        "unitary",
        ir.Unitary,
        [TwoDimensionalMatrix, MultiTarget],
        {"input_type": float},
    ),
    (
        Gate.Unitary,
        "unitary",
        ir.Unitary,
        [TwoDimensionalMatrix, MultiTarget],
        {"input_type": int},
    ),
]

parameterizable_gates = [
    Gate.GPhase,
    Gate.Rx,
    Gate.Ry,
    Gate.Rz,
    Gate.U,
    Gate.PhaseShift,
    Gate.PSwap,
    Gate.XX,
    Gate.XY,
    Gate.YY,
    Gate.ZZ,
    Gate.CPhaseShift,
    Gate.CPhaseShift00,
    Gate.CPhaseShift01,
    Gate.CPhaseShift10,
    Gate.GPi,
    Gate.GPi2,
    Gate.PRx,
    Gate.MS,
]


invalid_unitary_matrices = [
    (np.array([[1]])),
    (np.array([1])),
    (np.array([0, 1, 2])),
    (np.array([[0, 1], [1, 2], [3, 4]])),
    (np.array([[0, 1, 2], [2, 3]], dtype=object)),
    (np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])),
    (np.array([[0, 1], [1, 1]])),
]


def no_target_valid_input(**kwargs):
    return {}


def single_target_valid_input(**kwargs):
    return {"target": 2}


def double_target_valid_ir_input(**kwargs):
    return {"targets": [2, 3]}


def double_target_valid_input(**kwargs):
    return {"target1": 2, "target2": 3}


def angle_valid_input(**kwargs):
    return {"angle": 0.123}


def double_angle_valid_input(**kwargs):
    return {"angle_1": 0.123, "angle_2": 3.567}


def triple_angle_valid_input(**kwargs):
    return {"angle_1": 0.123, "angle_2": 4.567, "angle_3": 8.910}


def single_control_valid_input(**kwargs):
    return {"control": 0}


def single_neg_control_valid_input(**kwargs):
    return {"control": [0], "control_state": [0]}


def double_control_valid_ir_input(**kwargs):
    return {"controls": [0, 1]}


def double_control_valid_input(**kwargs):
    return {"control1": 0, "control2": 1}


def multi_target_valid_input(**kwargs):
    return {"targets": [5]}


def two_dimensional_matrix_valid_ir_input(**kwargs):
    return {"matrix": [[[0, 0], [1, 0]], [[1, 0], [0, 0]]]}


def two_dimensional_matrix_valid_input(**kwargs):
    input_type = kwargs.get("input_type")
    return {"matrix": np.array([[input_type(0), input_type(1)], [input_type(1), input_type(0)]])}


valid_ir_switcher = {
    "NoTarget": no_target_valid_input,
    "SingleTarget": single_target_valid_input,
    "DoubleTarget": double_target_valid_ir_input,
    "Angle": angle_valid_input,
    "DoubleAngle": double_angle_valid_input,
    "TripleAngle": triple_angle_valid_input,
    "SingleControl": single_control_valid_input,
    "SingleNegControlModifier": single_neg_control_valid_input,
    "DoubleControl": double_control_valid_ir_input,
    "MultiTarget": multi_target_valid_input,
    "TwoDimensionalMatrix": two_dimensional_matrix_valid_ir_input,
}

valid_subroutine_switcher = dict(
    valid_ir_switcher,
    **{
        "TwoDimensionalMatrix": two_dimensional_matrix_valid_input,
        "DoubleTarget": double_target_valid_input,
        "DoubleControl": double_control_valid_input,
    },
)


def create_valid_ir_input(irsubclasses):
    input = {}
    for subclass in irsubclasses:
        input |= valid_ir_switcher.get(subclass.__name__, lambda: "Invalid subclass")()
    return input


def create_valid_subroutine_input(irsubclasses, **kwargs):
    input = {}
    for subclass in irsubclasses:
        input |= valid_subroutine_switcher.get(subclass.__name__, lambda: "Invalid subclass")(
            **kwargs
        )
    return input


def create_valid_target_input(irsubclasses):
    qubit_set = []
    control_qubit_set = []
    control_state = None
    # based on the concept that control goes first in target input
    for subclass in irsubclasses:
        if subclass == NoTarget:
            qubit_set.extend(list(no_target_valid_input().values()))
        elif subclass == SingleTarget:
            qubit_set.extend(list(single_target_valid_input().values()))
        elif subclass == DoubleTarget:
            qubit_set.extend(list(double_target_valid_ir_input().values()))
        elif subclass == MultiTarget:
            qubit_set.extend(list(multi_target_valid_input().values()))
        elif subclass == SingleControl:
            qubit_set = list(single_control_valid_input().values()) + qubit_set
        elif subclass == SingleNegControlModifier:
            control_qubit_set = list(single_neg_control_valid_input()["control"])
            control_state = list(single_neg_control_valid_input()["control_state"])
        elif subclass == DoubleControl:
            qubit_set = list(double_control_valid_ir_input().values()) + qubit_set
        elif subclass not in (Angle, TwoDimensionalMatrix, DoubleAngle, TripleAngle):
            raise ValueError("Invalid subclass")
    input = {"target": QubitSet(qubit_set)}
    input["control"] = QubitSet(control_qubit_set)
    input["control_state"] = control_state
    return input


def create_valid_gate_class_input(irsubclasses, **kwargs):
    input = {}
    if Angle in irsubclasses:
        input.update(angle_valid_input())
    if DoubleAngle in irsubclasses:
        input.update(double_angle_valid_input())
    if TripleAngle in irsubclasses:
        input.update(triple_angle_valid_input())
    if TwoDimensionalMatrix in irsubclasses:
        input.update(two_dimensional_matrix_valid_input(**kwargs))
    return input


def create_valid_instruction_input(testclass, irsubclasses, **kwargs):
    input = create_valid_target_input(irsubclasses)
    input["operator"] = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
    return input


def calculate_qubit_count(irsubclasses):
    qubit_count = 0
    for subclass in irsubclasses:
        if subclass == SingleTarget:
            qubit_count += 1
        elif subclass == DoubleTarget:
            qubit_count += 2
        elif subclass == SingleControl:
            qubit_count += 1
        elif subclass == DoubleControl:
            qubit_count += 2
        elif subclass == MultiTarget:
            qubit_count += 3
        elif subclass not in (
            NoTarget,
            Angle,
            TwoDimensionalMatrix,
            DoubleAngle,
            TripleAngle,
        ):
            raise ValueError("Invalid subclass")
    return qubit_count


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_ir_gate_level(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    if irclass is not None:
        expected = irclass(**create_valid_ir_input(irsubclasses))
        actual = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs)).to_ir(
            **create_valid_target_input(irsubclasses)
        )
        assert actual == expected


@pytest.mark.parametrize(
    "gate, target, serialization_properties, expected_ir",
    [
        (
            Gate.Rx(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "rx(0.17) q[4];",
        ),
        (
            Gate.Rx(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "rx(0.17) $4;",
        ),
        (
            Gate.X(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "x q[4];",
        ),
        (
            Gate.X(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "x $4;",
        ),
        (
            Gate.Z(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "z q[4];",
        ),
        (
            Gate.Z(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "z $4;",
        ),
        (
            Gate.Y(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "y q[4];",
        ),
        (
            Gate.Y(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "y $4;",
        ),
        (
            Gate.H(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "h q[4];",
        ),
        (
            Gate.H(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "h $4;",
        ),
        (
            Gate.Ry(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "ry(0.17) q[4];",
        ),
        (
            Gate.Ry(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "ry(0.17) $4;",
        ),
        (
            Gate.ZZ(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "zz(0.17) q[4], q[5];",
        ),
        (
            Gate.ZZ(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "zz(0.17) $4, $5;",
        ),
        (
            Gate.I(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "i q[4];",
        ),
        (
            Gate.I(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "i $4;",
        ),
        (
            Gate.V(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "v q[4];",
        ),
        (
            Gate.V(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "v $4;",
        ),
        (
            Gate.CY(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cy q[0], q[1];",
        ),
        (
            Gate.CY(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cy $0, $1;",
        ),
        (
            Gate.Rz(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "rz(0.17) q[4];",
        ),
        (
            Gate.Rz(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "rz(0.17) $4;",
        ),
        (
            Gate.U(angle_1=0.17, angle_2=3.45, angle_3=5.21),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "U(0.17, 3.45, 5.21) q[4];",
        ),
        (
            Gate.U(angle_1=0.17, angle_2=3.45, angle_3=5.21),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "U(0.17, 3.45, 5.21) $4;",
        ),
        (
            Gate.XX(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "xx(0.17) q[4], q[5];",
        ),
        (
            Gate.XX(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "xx(0.17) $4, $5;",
        ),
        (
            Gate.T(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "t q[4];",
        ),
        (
            Gate.T(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "t $4;",
        ),
        (
            Gate.CZ(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cz $0, $1;",
        ),
        (
            Gate.CZ(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cz q[0], q[1];",
        ),
        (
            Gate.YY(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "yy(0.17) q[4], q[5];",
        ),
        (
            Gate.YY(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "yy(0.17) $4, $5;",
        ),
        (
            Gate.XY(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "xy(0.17) q[4], q[5];",
        ),
        (
            Gate.XY(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "xy(0.17) $4, $5;",
        ),
        (
            Gate.ISwap(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "iswap $0, $1;",
        ),
        (
            Gate.ISwap(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "iswap q[0], q[1];",
        ),
        (
            Gate.Swap(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "swap $0, $1;",
        ),
        (
            Gate.Swap(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "swap q[0], q[1];",
        ),
        (
            Gate.ECR(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "ecr $0, $1;",
        ),
        (
            Gate.ECR(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "ecr q[0], q[1];",
        ),
        (
            Gate.CV(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cv $0, $1;",
        ),
        (
            Gate.CV(),
            [0, 1],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cv q[0], q[1];",
        ),
        (
            Gate.Vi(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "vi q[4];",
        ),
        (
            Gate.Vi(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "vi $4;",
        ),
        (
            Gate.CSwap(),
            [0, 1, 2],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cswap q[0], q[1], q[2];",
        ),
        (
            Gate.CSwap(),
            [0, 1, 2],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cswap $0, $1, $2;",
        ),
        (
            Gate.CPhaseShift01(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cphaseshift01(0.17) q[4], q[5];",
        ),
        (
            Gate.CPhaseShift01(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cphaseshift01(0.17) $4, $5;",
        ),
        (
            Gate.CPhaseShift00(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cphaseshift00(0.17) q[4], q[5];",
        ),
        (
            Gate.CPhaseShift00(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cphaseshift00(0.17) $4, $5;",
        ),
        (
            Gate.CPhaseShift(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cphaseshift(0.17) q[4], q[5];",
        ),
        (
            Gate.CPhaseShift(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cphaseshift(0.17) $4, $5;",
        ),
        (
            Gate.S(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "s q[4];",
        ),
        (
            Gate.S(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "s $4;",
        ),
        (
            Gate.Si(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "si q[4];",
        ),
        (
            Gate.Si(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "si $4;",
        ),
        (
            Gate.Ti(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "ti q[4];",
        ),
        (
            Gate.Ti(),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "ti $4;",
        ),
        (
            Gate.PhaseShift(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "phaseshift(0.17) q[4];",
        ),
        (
            Gate.PhaseShift(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "phaseshift(0.17) $4;",
        ),
        (
            Gate.CNot(),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cnot q[4], q[5];",
        ),
        (
            Gate.CNot(),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cnot $4, $5;",
        ),
        (
            Gate.PSwap(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "pswap(0.17) q[4], q[5];",
        ),
        (
            Gate.PSwap(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "pswap(0.17) $4, $5;",
        ),
        (
            Gate.CPhaseShift10(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "cphaseshift10(0.17) q[4], q[5];",
        ),
        (
            Gate.CPhaseShift10(angle=0.17),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "cphaseshift10(0.17) $4, $5;",
        ),
        (
            Gate.CCNot(),
            [4, 5, 6],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "ccnot q[4], q[5], q[6];",
        ),
        (
            Gate.CCNot(),
            [4, 5, 6],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "ccnot $4, $5, $6;",
        ),
        (
            Gate.Unitary(Gate.CCNot().to_matrix()),
            [4, 5, 6],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket unitary(["
            "[1.0, 0, 0, 0, 0, 0, 0, 0], "
            "[0, 1.0, 0, 0, 0, 0, 0, 0], "
            "[0, 0, 1.0, 0, 0, 0, 0, 0], "
            "[0, 0, 0, 1.0, 0, 0, 0, 0], "
            "[0, 0, 0, 0, 1.0, 0, 0, 0], "
            "[0, 0, 0, 0, 0, 1.0, 0, 0], "
            "[0, 0, 0, 0, 0, 0, 0, 1.0], "
            "[0, 0, 0, 0, 0, 0, 1.0, 0]"
            "]) q[4], q[5], q[6]",
        ),
        (
            Gate.Unitary(Gate.CCNot().to_matrix()),
            [4, 5, 6],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "#pragma braket unitary(["
            "[1.0, 0, 0, 0, 0, 0, 0, 0], "
            "[0, 1.0, 0, 0, 0, 0, 0, 0], "
            "[0, 0, 1.0, 0, 0, 0, 0, 0], "
            "[0, 0, 0, 1.0, 0, 0, 0, 0], "
            "[0, 0, 0, 0, 1.0, 0, 0, 0], "
            "[0, 0, 0, 0, 0, 1.0, 0, 0], "
            "[0, 0, 0, 0, 0, 0, 0, 1.0], "
            "[0, 0, 0, 0, 0, 0, 1.0, 0]"
            "]) $4, $5, $6",
        ),
        (
            Gate.Unitary(np.round(Gate.ECR().to_matrix(), 8)),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket unitary(["
            "[0, 0, 0.70710678, 0.70710678im], "
            "[0, 0, 0.70710678im, 0.70710678], "
            "[0.70710678, -0.70710678im, 0, 0], "
            "[-0.70710678im, 0.70710678, 0, 0]"
            "]) q[4], q[5]",
        ),
        (
            Gate.Unitary(np.round(Gate.ECR().to_matrix(), 8)),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "#pragma braket unitary(["
            "[0, 0, 0.70710678, 0.70710678im], "
            "[0, 0, 0.70710678im, 0.70710678], "
            "[0.70710678, -0.70710678im, 0, 0], "
            "[-0.70710678im, 0.70710678, 0, 0]"
            "]) $4, $5",
        ),
        (
            Gate.Unitary(np.round(Gate.T().to_matrix(), 8)),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket unitary([[1.0, 0], [0, 0.70710678 + 0.70710678im]]) q[4]",
        ),
        (
            Gate.Unitary(np.round(Gate.T().to_matrix(), 8)),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "#pragma braket unitary([[1.0, 0], [0, 0.70710678 + 0.70710678im]]) $4",
        ),
        (
            Gate.Unitary(np.array([[1.0, 0], [0, 0.70710678 - 0.70710678j]])),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "#pragma braket unitary([[1.0, 0], [0, 0.70710678 - 0.70710678im]]) q[4]",
        ),
        (
            Gate.PulseGate(
                PulseSequence().play(
                    Frame("user_frame", Port("device_port_x", 1e-9), 1e9),
                    ArbitraryWaveform([1, 2], "arb_wf"),
                ),
                1,
            ),
            [0],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "\n".join(["cal {", "    play(user_frame, arb_wf);", "}"]),
        ),
        (
            Gate.GPi(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "gpi(0.17) q[4];",
        ),
        (
            Gate.GPi(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "gpi(0.17) $4;",
        ),
        (
            Gate.GPi2(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "gpi2(0.17) q[4];",
        ),
        (
            Gate.GPi2(angle=0.17),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "gpi2(0.17) $4;",
        ),
        (
            Gate.PRx(angle_1=0.17, angle_2=3.45),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            "prx(0.17, 3.45) q[4];",
        ),
        (
            Gate.PRx(angle_1=0.17, angle_2=3.45),
            [4],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            "prx(0.17, 3.45) $4;",
        ),
        (
            Gate.MS(angle_1=0.17, angle_2=3.45),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.VIRTUAL),
            f"ms(0.17, 3.45, {np.pi / 2}) q[4], q[5];",
        ),
        (
            Gate.MS(angle_1=0.17, angle_2=3.45),
            [4, 5],
            OpenQASMSerializationProperties(qubit_reference_type=QubitReferenceType.PHYSICAL),
            f"ms(0.17, 3.45, {np.pi / 2}) $4, $5;",
        ),
    ],
)
def test_gate_to_ir_openqasm(gate, target, serialization_properties, expected_ir):
    assert (
        gate.to_ir(
            target, ir_type=IRType.OPENQASM, serialization_properties=serialization_properties
        )
        == expected_ir
    )


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_ir_instruction_level(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    if irclass is not None:
        expected = irclass(**create_valid_ir_input(irsubclasses))
        instruction = Instruction(
            **create_valid_instruction_input(testclass, irsubclasses, **kwargs)
        )
        actual = instruction.to_ir()
        assert actual == expected


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_gate_subroutine(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    qubit_count = calculate_qubit_count(irsubclasses)
    subroutine = getattr(Circuit(), subroutine_name)
    assert subroutine(**create_valid_subroutine_input(irsubclasses, **kwargs)) == Circuit(
        Instruction(**create_valid_instruction_input(testclass, irsubclasses, **kwargs))
    )
    if qubit_count == 1:
        multi_targets = [0, 1, 2]
        instruction_list = [
            Instruction(
                operator=testclass(**create_valid_gate_class_input(irsubclasses, **kwargs)),
                target=target,
            )
            for target in multi_targets
        ]
        subroutine = getattr(Circuit(), subroutine_name)
        subroutine_input = {"target": multi_targets}
        if Angle in irsubclasses:
            subroutine_input.update(angle_valid_input())
        if DoubleAngle in irsubclasses:
            subroutine_input.update(double_angle_valid_input())
        if TripleAngle in irsubclasses:
            subroutine_input.update(triple_angle_valid_input())
        assert subroutine(**subroutine_input) == Circuit(instruction_list)


@pytest.mark.parametrize(
    "control, control_state, instruction_set",
    [
        (
            2,
            None,
            Instruction(**create_valid_instruction_input(Gate.PhaseShift, [SingleTarget, Angle])),
        ),
        (
            2,
            [0],
            [
                Instruction(operator=Gate.X(), target=2),
                Instruction(
                    **create_valid_instruction_input(Gate.PhaseShift, [SingleTarget, Angle])
                ),
                Instruction(operator=Gate.X(), target=2),
            ],
        ),
        (
            [0, 2],
            [0, 1],
            Instruction(
                **create_valid_instruction_input(
                    Gate.PhaseShift, [SingleTarget, SingleNegControlModifier, Angle]
                )
            ),
        ),
    ],
)
def test_control_gphase_subroutine(control, control_state, instruction_set):
    subroutine = getattr(Circuit(), "gphase")
    assert subroutine(angle=0.123, control=control, control_state=control_state) == Circuit(
        instruction_set
    )


def test_angle_gphase_is_none():
    with pytest.raises(ValueError, match="angle must not be None"):
        Gate.GPhase(angle=None)


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_gate_adjoint_expansion_correct(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    gate = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
    matrices = [elem.to_matrix() for elem in gate.adjoint()]
    matrices.append(gate.to_matrix())
    identity = np.eye(2**gate.qubit_count)
    assert np.allclose(functools.reduce(lambda a, b: a @ b, matrices), identity)


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_gate_to_matrix(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    gate1 = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
    gate2 = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
    assert isinstance(gate1.to_matrix(), np.ndarray)
    assert gate1.matrix_equivalence(gate2)


@pytest.mark.parametrize("testclass,subroutine_name,irclass,irsubclasses,kwargs", testdata)
def test_fixed_qubit_count(testclass, subroutine_name, irclass, irsubclasses, kwargs):
    fixed_qubit_count = testclass.fixed_qubit_count()
    if fixed_qubit_count is not NotImplemented:
        gate = testclass(**create_valid_gate_class_input(irsubclasses, **kwargs))
        assert gate.qubit_count == fixed_qubit_count


# Additional Unitary gate tests


def test_equality():
    u1 = Gate.Unitary(np.array([[0 + 0j, 1 + 0j], [1 + 0j, 0 + 0j]]))
    u2 = Gate.Unitary(np.array([[0, 1], [1, 0]], dtype=np.float32), display_name=["u2"])
    other_gate = Gate.Unitary(np.array([[1, 0], [0, 1]]))
    non_gate = "non gate"

    assert u1 == u2
    assert u1 is not u2
    assert u1 != other_gate
    assert u1 != non_gate


def test_free_param_equality():
    param1 = FreeParameter("theta")
    param2 = FreeParameter("phi")
    rx1 = Gate.Rx(param1)
    rx2 = Gate.Rx(param1)
    other_gate = Gate.Rx(param2)

    assert rx1 == rx2
    assert rx1 is not rx2
    assert rx1 != other_gate
    assert rx1 != param1


def test_large_unitary():
    matrix = np.eye(16, dtype=np.float32)
    # Permute rows of matrix
    matrix[[*range(16)]] = matrix[[(i + 1) % 16 for i in range(16)]]
    unitary = Gate.Unitary(matrix)
    assert unitary.qubit_count == 4


@pytest.mark.parametrize("gate", parameterizable_gates)
def test_bind_values(gate):
    double_angled = gate.__name__ in ["PRx"]
    triple_angled = gate.__name__ in ("MS", "U")
    num_params = 1
    if triple_angled:
        num_params = 3
    elif double_angled:
        num_params = 2
    thetas = [FreeParameter(f"theta_{i}") for i in range(num_params)]
    mapping = {f"theta_{i}": i for i in range(num_params)}
    param_gate = gate(*thetas)
    new_gate = param_gate.bind_values(**mapping)
    expected = gate(*range(num_params))

    assert type(new_gate) is type(param_gate) and new_gate == expected
    if triple_angled:
        for angle in new_gate.angle_1, new_gate.angle_2, new_gate.angle_3:
            assert isinstance(angle, float)
    elif double_angled:
        for angle in new_gate.angle_1, new_gate.angle_2:
            assert isinstance(angle, float)
    else:
        assert isinstance(new_gate.angle, float)


def test_bind_values_pulse_gate():
    qubit_count = 1
    frame = Frame("user_frame", Port("device_port_x", 1e-9), 1e9)
    gate = Gate.PulseGate(
        PulseSequence()
        .set_frequency(frame, FreeParameter("a") + FreeParameter("c"))
        .delay(frame, FreeParameter("d")),
        qubit_count,
    )

    def to_ir(pulse_gate):
        return pulse_gate.to_ir(range(pulse_gate.qubit_count), IRType.OPENQASM)

    a = 3
    a_bound = gate.bind_values(a=a)
    a_bound_ir = to_ir(a_bound)

    assert a_bound_ir == "\n".join([
        "cal {",
        "    set_frequency(user_frame, 3.0 + c);",
        "    delay[d * 1s] user_frame;",
        "}",
    ])

    assert a_bound_ir == to_ir(
        Gate.PulseGate(gate.pulse_sequence.make_bound_pulse_sequence({"a": a}), qubit_count)
    )
    assert a_bound_ir != to_ir(gate)

    c = 4e-6
    ac_bound = a_bound.bind_values(c=c)
    ac_bound_ir = to_ir(ac_bound)
    assert ac_bound_ir == to_ir(
        Gate.PulseGate(a_bound.pulse_sequence.make_bound_pulse_sequence({"c": c}), qubit_count)
    )
    assert ac_bound_ir != a_bound_ir


def test_pulse_gate_capture_throws():
    with pytest.raises(ValueError):
        Circuit().pulse_gate(
            0, PulseSequence().capture_v0(Frame("user_frame", Port("device_port_x", dt=1e-9), 1e9))
        )


@pytest.mark.parametrize("matrix", invalid_unitary_matrices)
def test_unitary_invalid_matrix(matrix):
    with pytest.raises(ValueError):
        Gate.Unitary(matrix=matrix)


def test_unitary_matrix_target_size_mismatch():
    with pytest.raises(ValueError):
        Circuit().unitary(
            matrix=np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]), targets=[0]
        )


def test_pulse_gate_to_matrix():
    with pytest.raises(NotImplementedError):
        Gate.PulseGate(
            PulseSequence().play(
                Frame("user_frame", Port("device_port_x", 1e-9), 1e9),
                ArbitraryWaveform([1, 2], "arb_wf"),
            ),
            1,
        ).to_matrix()


@pytest.mark.parametrize(
    "gate, target, control, control_state, expected_ir",
    (
        (Gate.H(), QubitSet(0), QubitSet(1), None, "ctrl @ h q[1], q[0];"),
        (Gate.H(), QubitSet(0), QubitSet([1, 2]), None, "ctrl(2) @ h q[1], q[2], q[0];"),
        (Gate.Ry(angle=1.23), QubitSet(0), QubitSet([2]), None, "ctrl @ ry(1.23) q[2], q[0];"),
        (
            Gate.MS(angle_1=0.17, angle_2=3.45),
            QubitSet(0),
            QubitSet([1, 2]),
            None,
            f"ctrl(2) @ ms(0.17, 3.45, {np.pi / 2}) q[1], q[2], q[0];",
        ),
        (
            Gate.CCNot(),
            QubitSet([0, 1, 2]),
            QubitSet([3, 4]),
            None,
            "ctrl(2) @ ccnot q[3], q[4], q[0], q[1], q[2];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3, 4, 5, 6, 7]),
            [1, 1, 1, 0, 0, 1, 0],
            "ctrl(3) @ negctrl(2) @ ctrl @ negctrl @ "
            "z q[1], q[2], q[3], q[4], q[5], q[6], q[7], q[0];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3, 4, 5, 6, 7]),
            "1110010",
            "ctrl(3) @ negctrl(2) @ ctrl @ negctrl @ "
            "z q[1], q[2], q[3], q[4], q[5], q[6], q[7], q[0];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3, 4, 5, 6, 7]),
            114,
            "ctrl(3) @ negctrl(2) @ ctrl @ negctrl @ "
            "z q[1], q[2], q[3], q[4], q[5], q[6], q[7], q[0];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3]),
            [1, 0],
            "negctrl @ ctrl @ negctrl @ z q[1], q[2], q[3], q[0];",
        ),
        (
            Gate.Z(),
            QubitSet([0]),
            QubitSet([1, 2, 3]),
            "10",
            "negctrl @ ctrl @ negctrl @ z q[1], q[2], q[3], q[0];",
        ),
        (
            Gate.GPhase(0.3),
            QubitSet([]),
            QubitSet([1]),
            "1",
            "ctrl @ gphase(0.3) q[1];",
        ),
    ),
)
def test_gate_control(gate, target, control, control_state, expected_ir):
    serialization_properties = OpenQASMSerializationProperties(
        qubit_reference_type=QubitReferenceType.VIRTUAL
    )
    assert (
        gate.to_ir(
            target,
            control=control,
            control_state=control_state,
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
        )
        == expected_ir
    )


@pytest.mark.parametrize(
    "control, control_state, error_string",
    (
        (
            [0, 1],
            [1, 0, 1],
            "State value represents a binary sequence of length greater "
            "than the specified number of qubits.",
        ),
        (
            [0, 1],
            "101",
            "State value represents a binary sequence of length greater "
            "than the specified number of qubits.",
        ),
        (
            [0, 1],
            5,
            "State value represents a binary sequence of length greater "
            "than the specified number of qubits.",
        ),
    ),
)
def test_gate_control_invalid_state(control, control_state, error_string):
    with pytest.raises(ValueError, match=error_string):
        Gate.X().to_ir(
            target=[0],
            control=control,
            control_state=control_state,
            ir_type=IRType.OPENQASM,
            serialization_properties=OpenQASMSerializationProperties(
                qubit_reference_type=QubitReferenceType.VIRTUAL
            ),
        )


@pytest.mark.parametrize(
    "gate, target, power, expected_ir",
    (
        (Gate.H(), QubitSet(0), 2, "pow(2) @ h q[0];"),
        (Gate.H(), QubitSet(0), 2.0, "pow(2.0) @ h q[0];"),
        (Gate.H(), QubitSet(0), 2.5, "pow(2.5) @ h q[0];"),
        (Gate.H(), QubitSet(0), 0, "pow(0) @ h q[0];"),
        (Gate.H(), QubitSet(0), -2, "inv @ pow(2) @ h q[0];"),
        (Gate.H(), QubitSet(0), -2.0, "inv @ pow(2.0) @ h q[0];"),
        (Gate.H(), QubitSet(0), -2.5, "inv @ pow(2.5) @ h q[0];"),
    ),
)
def test_gate_power(gate, target, power, expected_ir):
    serialization_properties = OpenQASMSerializationProperties(
        qubit_reference_type=QubitReferenceType.VIRTUAL
    )
    assert (
        gate.to_ir(
            target,
            power=power,
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
        )
        == expected_ir
    )


def test_hash():
    assert hash(Gate.Unitary(Gate.CCNot().to_matrix())) == hash(
        Gate.Unitary(Gate.CCNot().to_matrix())
    )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from numbers import Number

import numpy as np

from braket.ahs.discretization_types import DiscretizationError, DiscretizationProperties


class SiteType(Enum):
    VACANT = "Vacant"
    FILLED = "Filled"


@dataclass
class AtomArrangementItem:
    """Represents an item (coordinate and metadata) in an atom arrangement."""

    coordinate: tuple[Number, Number]
    site_type: SiteType

    def _validate_coordinate(self) -> None:
        if len(self.coordinate) != 2:
            raise ValueError(f"{self.coordinate} must be of length 2")
        for idx, num in enumerate(self.coordinate):
            if not isinstance(num, Number):
                raise TypeError(f"{num} at position {idx} must be a number")

    def _validate_site_type(self) -> None:
        allowed_site_types = {SiteType.FILLED, SiteType.VACANT}
        if self.site_type not in allowed_site_types:
            raise ValueError(f"{self.site_type} must be one of {allowed_site_types}")

    def __post_init__(self) -> None:
        self._validate_coordinate()
        self._validate_site_type()


class AtomArrangement:
    def __init__(self):
        """Represents a set of coordinates that can be used as a register to an
        AnalogHamiltonianSimulation.
        """
        self._sites = []

    def add(
        self,
        coordinate: tuple[Number, Number] | np.ndarray,
        site_type: SiteType = SiteType.FILLED,
    ) -> AtomArrangement:
        """Add a coordinate to the atom arrangement.

        Args:
            coordinate (Union[tuple[Number, Number], ndarray]): The coordinate of the
                atom (in meters). The coordinates can be a numpy array of shape (2,)
                or a tuple of int, float, Decimal
            site_type (SiteType): The type of site. Optional. Default is FILLED.

        Returns:
            AtomArrangement: returns self (to allow for chaining).
        """
        self._sites.append(AtomArrangementItem(tuple(coordinate), site_type))
        return self

    def coordinate_list(self, coordinate_index: Number) -> list[Number]:
        """Returns all the coordinates at the given index.

        Args:
            coordinate_index (Number): The index to get for each coordinate.

        Returns:
            list[Number]:The list of coordinates at the given index.

        Example:
            To get a list of all x-coordinates: coordinate_list(0)
            To get a list of all y-coordinates: coordinate_list(1)
        """
        return [site.coordinate[coordinate_index] for site in self._sites]

    def __iter__(self) -> Iterator:
        return self._sites.__iter__()

    def __len__(self):
        return self._sites.__len__()

    def discretize(self, properties: DiscretizationProperties) -> AtomArrangement:
        """Creates a discretized version of the atom arrangement,
        rounding all site coordinates to the closest multiple of the
        resolution. The types of the sites are unchanged.

        Args:
            properties (DiscretizationProperties): Capabilities of a device that represent the
                resolution with which the device can implement the parameters.

        Raises:
            DiscretizationError: If unable to discretize the program.

        Returns:
            AtomArrangement: A new discretized atom arrangement.
        """
        try:
            position_res = properties.lattice.geometry.positionResolution
            discretized_arrangement = AtomArrangement()
            for site in self._sites:
                new_coordinates = tuple(
                    round(Decimal(c) / position_res) * position_res for c in site.coordinate
                )
                discretized_arrangement.add(new_coordinates, site.site_type)
        except Exception as e:
            raise DiscretizationError(f"Failed to discretize register {e}") from e
        else:
            return discretized_arrangement



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import os
import warnings
from contextlib import AbstractContextManager

from braket.aws.aws_device import AwsDevice
from braket.devices import Device


class DirectReservation(AbstractContextManager):
    """
    Context manager that modifies AwsQuantumTasks created within the context to use a reservation
    ARN for all tasks targeting the specified device. Note: this context manager only allows for
    one reservation at a time.

    Reservations are AWS account and device specific. Only the AWS account that created the
    reservation can use your reservation ARN. Additionally, the reservation ARN is only valid on the
    reserved device at the chosen start and end times.

    Args:
        device (Device | str | None): The Braket device for which you have a reservation ARN, or
            optionally the device ARN.
        reservation_arn (str | None): The Braket Direct reservation ARN to be applied to all
            quantum tasks run within the context.

    Examples:
        As a context manager
        >>> with DirectReservation(device_arn, reservation_arn="<my_reservation_arn>"):
        ...     task1 = device.run(circuit, shots)
        ...     task2 = device.run(circuit, shots)

        or start the reservation
        >>> DirectReservation(device_arn, reservation_arn="<my_reservation_arn>").start()
        ... task1 = device.run(circuit, shots)
        ... task2 = device.run(circuit, shots)

    References:

    [1] https://docs.aws.amazon.com/braket/latest/developerguide/braket-reservations.html
    """

    _is_active = False  # Class variable to track active reservation context

    def __init__(self, device: Device | str | None, reservation_arn: str | None):
        if isinstance(device, AwsDevice):
            self.device_arn = device.arn
        elif isinstance(device, str):
            self.device_arn = AwsDevice(device).arn  # validate ARN early
        elif isinstance(device, Device) or device is None:  # LocalSimulator
            warnings.warn(
                "Using a local simulator with the reservation. For a reservation on a QPU, please "
                "ensure the device matches the reserved Braket device.",
                stacklevel=2,
            )
            self.device_arn = ""  # instead of None, use empty string
        else:
            raise TypeError("Device must be an AwsDevice or its ARN, or a local simulator device.")

        self.reservation_arn = reservation_arn

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # noqa: ANN001
        self.stop()

    def start(self) -> None:
        """Start the reservation context."""
        if DirectReservation._is_active:
            raise RuntimeError("Another reservation is already active.")

        os.environ["AMZN_BRAKET_RESERVATION_DEVICE_ARN"] = self.device_arn
        if self.reservation_arn:
            os.environ["AMZN_BRAKET_RESERVATION_TIME_WINDOW_ARN"] = self.reservation_arn
        DirectReservation._is_active = True

    def stop(self) -> None:
        """Stop the reservation context."""
        if not DirectReservation._is_active:
            warnings.warn("Reservation context is not active.", stacklevel=2)
            return
        os.environ.pop("AMZN_BRAKET_RESERVATION_DEVICE_ARN", None)
        os.environ.pop("AMZN_BRAKET_RESERVATION_TIME_WINDOW_ARN", None)
        DirectReservation._is_active = False



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
from unittest.mock import MagicMock, patch

import pytest

from braket.aws import AwsDevice, AwsSession, DirectReservation
from braket.devices import LocalSimulator

RESERVATION_ARN = "arn:aws:braket:us-east-1:123456789:reservation/uuid"
DEVICE_ARN = "arn:aws:braket:us-east-1:123456789:device/qpu/ionq/Forte-1"
VALUE_ERROR_MESSAGE = "Device must be an AwsDevice or its ARN, or a local simulator device."
RUNTIME_ERROR_MESSAGE = "Another reservation is already active."


@pytest.fixture
def aws_device():
    mock_device = MagicMock(spec=AwsDevice)
    mock_device._arn = DEVICE_ARN
    type(mock_device).arn = property(lambda x: DEVICE_ARN)
    return mock_device


def test_direct_reservation_aws_device(aws_device):
    with DirectReservation(aws_device, RESERVATION_ARN) as reservation:
        assert reservation.device_arn == DEVICE_ARN
        assert reservation.reservation_arn == RESERVATION_ARN
        assert reservation._is_active


def test_direct_reservation_device_str(aws_device):
    with patch(
        "braket.aws.AwsDevice.__init__",
        side_effect=lambda self, *args, **kwargs: setattr(self, "_arn", DEVICE_ARN),
        autospec=True,
    ):
        with patch("braket.aws.AwsDevice", return_value=aws_device, autospec=True):
            with DirectReservation(DEVICE_ARN, RESERVATION_ARN) as reservation:
                assert reservation.device_arn == DEVICE_ARN
                assert reservation.reservation_arn == RESERVATION_ARN
                assert reservation._is_active


def test_direct_reservation_local_simulator():
    mock_device = MagicMock(spec=LocalSimulator)
    with pytest.warns(UserWarning):
        with DirectReservation(mock_device, RESERVATION_ARN) as reservation:
            assert os.environ["AMZN_BRAKET_RESERVATION_DEVICE_ARN"] == ""
            assert os.environ["AMZN_BRAKET_RESERVATION_TIME_WINDOW_ARN"] == RESERVATION_ARN
            assert reservation._is_active is True


@pytest.mark.parametrize("device", [123, False, [aws_device], {"a": 1}])
def test_direct_reservation_invalid_inputs(device):
    with pytest.raises(TypeError):
        DirectReservation(device, RESERVATION_ARN)


def test_direct_reservation_local_no_reservation():
    mock_device = MagicMock(spec=LocalSimulator)
    mock_device.create_quantum_task = MagicMock()
    kwargs = {
        "program": {"ir": '{"instructions":[]}', "qubitCount": 4},
        "shots": 1,
    }
    with DirectReservation(mock_device, None):
        mock_device.create_quantum_task(**kwargs)
        mock_device.create_quantum_task.assert_called_once_with(**kwargs)


def test_context_management(aws_device):
    with DirectReservation(aws_device, RESERVATION_ARN):
        assert os.getenv("AMZN_BRAKET_RESERVATION_DEVICE_ARN") == DEVICE_ARN
        assert os.getenv("AMZN_BRAKET_RESERVATION_TIME_WINDOW_ARN") == RESERVATION_ARN
    assert not os.getenv("AMZN_BRAKET_RESERVATION_DEVICE_ARN")
    assert not os.getenv("AMZN_BRAKET_RESERVATION_TIME_WINDOW_ARN")


def test_start_reservation_already_active(aws_device):
    reservation = DirectReservation(aws_device, RESERVATION_ARN)
    reservation.start()
    with pytest.raises(RuntimeError, match=RUNTIME_ERROR_MESSAGE):
        reservation.start()
    reservation.stop()


def test_stop_reservation_not_active(aws_device):
    reservation = DirectReservation(aws_device, RESERVATION_ARN)
    with pytest.warns(UserWarning):
        reservation.stop()


def test_multiple_start_stop_cycles(aws_device):
    reservation = DirectReservation(aws_device, RESERVATION_ARN)
    reservation.start()
    reservation.stop()
    reservation.start()
    reservation.stop()
    assert not os.getenv("AMZN_BRAKET_RESERVATION_DEVICE_ARN")
    assert not os.getenv("AMZN_BRAKET_RESERVATION_TIME_WINDOW_ARN")


def test_two_direct_reservations(aws_device):
    with pytest.raises(RuntimeError, match=RUNTIME_ERROR_MESSAGE):
        with DirectReservation(aws_device, RESERVATION_ARN):
            with DirectReservation(aws_device, "reservation_arn_example_2"):
                pass


def test_create_quantum_task_with_correct_device_and_reservation(aws_device):
    kwargs = {"deviceArn": DEVICE_ARN, "shots": 1}
    with patch("boto3.client"):
        mock_client = MagicMock()
        aws_session = AwsSession(braket_client=mock_client)
        with DirectReservation(aws_device, RESERVATION_ARN):
            aws_session.create_quantum_task(**kwargs)
            kwargs["associations"] = [
                {
                    "arn": RESERVATION_ARN,
                    "type": "RESERVATION_TIME_WINDOW_ARN",
                }
            ]
            mock_client.create_quantum_task.assert_called_once_with(**kwargs)


def test_warning_for_overridden_reservation_arn(aws_device):
    kwargs = {
        "deviceArn": DEVICE_ARN,
        "shots": 1,
        "associations": [
            {
                "arn": "task_reservation_arn",
                "type": "RESERVATION_TIME_WINDOW_ARN",
            }
        ],
    }
    correct_kwargs = {
        "deviceArn": DEVICE_ARN,
        "shots": 1,
        "associations": [
            {
                "arn": RESERVATION_ARN,
                "type": "RESERVATION_TIME_WINDOW_ARN",
            }
        ],
    }
    with patch("boto3.client"):
        mock_client = MagicMock()
        aws_session = AwsSession(braket_client=mock_client)
        with pytest.warns(
            UserWarning,
            match="A reservation ARN was passed to 'CreateQuantumTask', but it is being overridden",
        ):
            with DirectReservation(aws_device, RESERVATION_ARN):
                aws_session.create_quantum_task(**kwargs)
                mock_client.create_quantum_task.assert_called_once_with(**correct_kwargs)


def test_warning_not_triggered_wrong_association_type():
    kwargs = {
        "deviceArn": DEVICE_ARN,
        "shots": 1,
        "associations": [{"type": "OTHER_TYPE"}],
    }
    with patch("boto3.client"):
        mock_client = MagicMock()
        aws_session = AwsSession(braket_client=mock_client)
        aws_session.create_quantum_task(**kwargs)
        mock_client.create_quantum_task.assert_called_once_with(**kwargs)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import re

from braket.aws import AwsSession
from braket.jobs import Framework, retrieve_image


def decorator_python_version():
    aws_session = AwsSession()
    image_uri = retrieve_image(Framework.BASE, aws_session.region)
    tag = aws_session.get_full_image_tag(image_uri)
    major_version, minor_version = re.search(r"-py(\d)(\d+)-", tag).groups()
    return int(major_version), int(minor_version)
